use std::fmt::Display;

use serde::{ser::Impossible, Serialize};

use super::ToValError;

pub(crate) struct MapKeySerializer;

fn key_must_be_a_string() -> ToValError {
    ToValError::KeyMustBeAString
}

impl serde::Serializer for MapKeySerializer {
    type Ok = String;
    type Error = ToValError;

    type SerializeSeq = Impossible<String, ToValError>;
    type SerializeTuple = Impossible<String, ToValError>;
    type SerializeTupleStruct = Impossible<String, ToValError>;
    type SerializeTupleVariant = Impossible<String, ToValError>;
    type SerializeMap = Impossible<String, ToValError>;
    type SerializeStruct = Impossible<String, ToValError>;
    type SerializeStructVariant = Impossible<String, ToValError>;

    #[inline]
    fn serialize_unit_variant(
        self,
        _name: &'static str,
        _variant_index: u32,
        variant: &'static str,
    ) -> Result<String, ToValError> {
        Ok(variant.to_owned())
    }

    #[inline]
    fn serialize_newtype_struct<T>(
        self,
        _name: &'static str,
        value: &T,
    ) -> Result<String, ToValError>
    where
        T: ?Sized + Serialize,
    {
        value.serialize(self)
    }

    fn serialize_bool(self, _value: bool) -> Result<String, ToValError> {
        Err(key_must_be_a_string())
    }

    fn serialize_i8(self, value: i8) -> Result<String, ToValError> {
        Ok(value.to_string())
    }

    fn serialize_i16(self, value: i16) -> Result<String, ToValError> {
        Ok(value.to_string())
    }

    fn serialize_i32(self, value: i32) -> Result<String, ToValError> {
        Ok(value.to_string())
    }

    fn serialize_i64(self, value: i64) -> Result<String, ToValError> {
        Ok(value.to_string())
    }

    fn serialize_u8(self, value: u8) -> Result<String, ToValError> {
        Ok(value.to_string())
    }

    fn serialize_u16(self, value: u16) -> Result<String, ToValError> {
        Ok(value.to_string())
    }

    fn serialize_u32(self, value: u32) -> Result<String, ToValError> {
        Ok(value.to_string())
    }

    fn serialize_u64(self, value: u64) -> Result<String, ToValError> {
        Ok(value.to_string())
    }

    fn serialize_f32(self, _value: f32) -> Result<String, ToValError> {
        Err(key_must_be_a_string())
    }

    fn serialize_f64(self, _value: f64) -> Result<String, ToValError> {
        Err(key_must_be_a_string())
    }

    #[inline]
    fn serialize_char(self, value: char) -> Result<String, ToValError> {
        Ok({
            let mut s = String::new();
            s.push(value);
            s
        })
    }

    #[inline]
    fn serialize_str(self, value: &str) -> Result<String, ToValError> {
        Ok(value.to_owned())
    }

    fn serialize_bytes(self, _value: &[u8]) -> Result<String, ToValError> {
        Err(key_must_be_a_string())
    }

    fn serialize_unit(self) -> Result<String, ToValError> {
        Err(key_must_be_a_string())
    }

    fn serialize_unit_struct(self, _name: &'static str) -> Result<String, ToValError> {
        Err(key_must_be_a_string())
    }

    fn serialize_newtype_variant<T>(
        self,
        _name: &'static str,
        _variant_index: u32,
        _variant: &'static str,
        _value: &T,
    ) -> Result<String, ToValError>
    where
        T: ?Sized + Serialize,
    {
        Err(key_must_be_a_string())
    }

    fn serialize_none(self) -> Result<String, ToValError> {
        Err(key_must_be_a_string())
    }

    fn serialize_some<T>(self, _value: &T) -> Result<String, ToValError>
    where
        T: ?Sized + Serialize,
    {
        Err(key_must_be_a_string())
    }

    fn serialize_seq(self, _len: Option<usize>) -> Result<Self::SerializeSeq, ToValError> {
        Err(key_must_be_a_string())
    }

    fn serialize_tuple(self, _len: usize) -> Result<Self::SerializeTuple, ToValError> {
        Err(key_must_be_a_string())
    }

    fn serialize_tuple_struct(
        self,
        _name: &'static str,
        _len: usize,
    ) -> Result<Self::SerializeTupleStruct, ToValError> {
        Err(key_must_be_a_string())
    }

    fn serialize_tuple_variant(
        self,
        _name: &'static str,
        _variant_index: u32,
        _variant: &'static str,
        _len: usize,
    ) -> Result<Self::SerializeTupleVariant, ToValError> {
        Err(key_must_be_a_string())
    }

    fn serialize_map(self, _len: Option<usize>) -> Result<Self::SerializeMap, ToValError> {
        Err(key_must_be_a_string())
    }

    fn serialize_struct(
        self,
        _name: &'static str,
        _len: usize,
    ) -> Result<Self::SerializeStruct, ToValError> {
        Err(key_must_be_a_string())
    }

    fn serialize_struct_variant(
        self,
        _name: &'static str,
        _variant_index: u32,
        _variant: &'static str,
        _len: usize,
    ) -> Result<Self::SerializeStructVariant, ToValError> {
        Err(key_must_be_a_string())
    }

    fn collect_str<T>(self, value: &T) -> Result<String, ToValError>
    where
        T: ?Sized + Display,
    {
        Ok(value.to_string())
    }
}
