use std::fmt::Display;

use ordered_float::NotNan;
use serde::{de::Visitor, Deserialize, Serialize, Serializer};
use thiserror::Error;

use super::{key::Key, key_e::KeyE, slim_boxed_slice::SlimBoxedSlice, Map, Val, ValE};

pub fn to_val<T: Serialize>(value: T) -> Result<Val, <ToValSerializer as Serializer>::Error> {
    value.serialize(ToValSerializer {})
}

pub struct ToValSerializer;

impl Serializer for ToValSerializer {
    type Ok = Val;

    type Error = ToValError;

    type SerializeSeq = SerializeVec;
    type SerializeTuple = SerializeVec;
    type SerializeTupleStruct = SerializeVec;
    type SerializeTupleVariant = SerializeTupleVariant;
    type SerializeMap = SerializeMap;
    type SerializeStruct = SerializeMap;
    type SerializeStructVariant = SerializeStructVariant;

    fn serialize_bool(self, v: bool) -> Result<Self::Ok, Self::Error> {
        Ok(ValE::Bool(v).into())
    }

    fn serialize_i8(self, v: i8) -> Result<Self::Ok, Self::Error> {
        self.serialize_i64(v as i64)
    }

    fn serialize_i16(self, v: i16) -> Result<Self::Ok, Self::Error> {
        self.serialize_i64(v as i64)
    }

    fn serialize_i32(self, v: i32) -> Result<Self::Ok, Self::Error> {
        self.serialize_i64(v as i64)
    }

    fn serialize_i64(self, v: i64) -> Result<Self::Ok, Self::Error> {
        self.serialize_f64(v as f64)
    }

    #[cfg(feature = "arbitrary_precision")]
    fn serialize_i128(self, value: i128) -> Result<Value> {
        Ok(ValE::Number(v.into()).into())
    }

    fn serialize_u8(self, v: u8) -> Result<Self::Ok, Self::Error> {
        self.serialize_u64(v as u64)
    }

    fn serialize_u16(self, v: u16) -> Result<Self::Ok, Self::Error> {
        self.serialize_u64(v as u64)
    }

    fn serialize_u32(self, v: u32) -> Result<Self::Ok, Self::Error> {
        self.serialize_u64(v as u64)
    }

    fn serialize_u64(self, v: u64) -> Result<Self::Ok, Self::Error> {
        self.serialize_f64(v as f64)
    }

    #[cfg(feature = "arbitrary_precision")]
    fn serialize_u128(self, value: u128) -> Result<Value> {
        Ok(ValE::Number(value.into()).into())
    }

    fn serialize_f32(self, v: f32) -> Result<Self::Ok, Self::Error> {
        self.serialize_f64(v as f64)
    }

    fn serialize_f64(self, v: f64) -> Result<Self::Ok, Self::Error> {
        // TODO: replace Nan and Infinity with null
        Ok(ValE::Number(NotNan::new(v).unwrap()).into())
    }

    fn serialize_char(self, v: char) -> Result<Self::Ok, Self::Error> {
        let mut buf = [0; 4];

        Ok(ValE::str(v.encode_utf8(&mut buf)).into())
    }

    fn serialize_str(self, v: &str) -> Result<Self::Ok, Self::Error> {
        Ok(ValE::str(v).into())
    }

    // TODO: special holder for bytes that is semantically equivalent to array?
    fn serialize_bytes(self, v: &[u8]) -> Result<Self::Ok, Self::Error> {
        let vec = v
            .iter()
            .map(|&b| ValE::Number(NotNan::new(b as f64).unwrap()).into())
            .collect();
        Ok(ValE::Array(SlimBoxedSlice::from_vec(vec)).into())
    }

    fn serialize_none(self) -> Result<Self::Ok, Self::Error> {
        self.serialize_unit()
    }

    fn serialize_some<T: ?Sized>(self, value: &T) -> Result<Self::Ok, Self::Error>
    where
        T: Serialize,
    {
        value.serialize(self)
    }

    fn serialize_unit(self) -> Result<Self::Ok, Self::Error> {
        Ok(ValE::Null.into())
    }

    fn serialize_unit_struct(self, _name: &'static str) -> Result<Self::Ok, Self::Error> {
        self.serialize_unit()
    }

    fn serialize_unit_variant(
        self,
        _name: &'static str,
        _variant_index: u32,
        variant: &'static str,
    ) -> Result<Self::Ok, Self::Error> {
        self.serialize_str(variant)
    }

    fn serialize_newtype_struct<T: ?Sized>(
        self,
        _name: &'static str,
        value: &T,
    ) -> Result<Self::Ok, Self::Error>
    where
        T: Serialize,
    {
        value.serialize(self)
    }

    fn serialize_newtype_variant<T: ?Sized>(
        self,
        _name: &'static str,
        _variant_index: u32,
        variant: &'static str,
        inner: &T,
    ) -> Result<Self::Ok, Self::Error>
    where
        T: Serialize,
    {
        let pair = vec![(KeyE::str(variant).into(), to_val(inner)?)];
        Ok(ValE::Object(Map::new(pair)).into())
    }

    fn serialize_seq(self, len: Option<usize>) -> Result<Self::SerializeSeq, Self::Error> {
        Ok(SerializeVec {
            // TODO: proper capacity from SlimBoxedSlice
            vec: Vec::with_capacity(len.unwrap_or(0)),
        })
    }

    fn serialize_tuple(self, len: usize) -> Result<Self::SerializeTuple, Self::Error> {
        self.serialize_seq(Some(len))
    }

    fn serialize_tuple_struct(
        self,
        _name: &'static str,
        len: usize,
    ) -> Result<Self::SerializeTupleStruct, Self::Error> {
        self.serialize_seq(Some(len))
    }

    fn serialize_tuple_variant(
        self,
        _name: &'static str,
        _variant_index: u32,
        variant: &'static str,
        len: usize,
    ) -> Result<Self::SerializeTupleVariant, Self::Error> {
        Ok(SerializeTupleVariant {
            name: String::from(variant),
            // TODO: proper capacity from SlimBoxedSlice
            vec: Vec::with_capacity(len),
        })
    }

    fn serialize_map(self, len: Option<usize>) -> Result<Self::SerializeMap, Self::Error> {
        Ok(SerializeMap {
            // TODO: proper capacity from SlimBoxedSlice
            map: Vec::with_capacity(len.unwrap_or(5)),
            next_key: None,
        })
    }

    fn serialize_struct(
        self,
        _name: &'static str,
        len: usize,
    ) -> Result<Self::SerializeStruct, Self::Error> {
        self.serialize_map(Some(len))
    }

    fn serialize_struct_variant(
        self,
        _name: &'static str,
        _variant_index: u32,
        _variant: &'static str,
        _len: usize,
    ) -> Result<Self::SerializeStructVariant, Self::Error> {
        todo!()
    }

    fn collect_str<T>(self, value: &T) -> Result<Self::Ok, Self::Error>
    where
        T: ?Sized + Display,
    {
        Ok(ValE::string(value.to_string()).into())
    }
}

#[derive(Error, Debug)]
pub enum ToValError {
    #[error("Key must be a string")]
    KeyMustBeAString,
    #[error("Other: {0}")]
    Other(String),
}

impl serde::ser::Error for ToValError {
    fn custom<T>(msg: T) -> Self
    where
        T: Display,
    {
        ToValError::Other(msg.to_string())
    }
}

pub struct SerializeVec {
    vec: Vec<Val>,
}

pub struct SerializeTupleVariant {
    name: String,
    vec: Vec<Val>,
}

pub struct SerializeMap {
    map: Vec<(Key, Val)>,
    next_key: Option<String>,
}

pub struct SerializeStructVariant {
    name: String,
    map: Vec<(Key, Val)>,
}

impl serde::ser::SerializeSeq for SerializeVec {
    type Ok = Val;
    type Error = ToValError;

    fn serialize_element<T>(&mut self, value: &T) -> Result<(), ToValError>
    where
        T: ?Sized + Serialize,
    {
        self.vec.push(to_val(value)?);
        Ok(())
    }

    fn end(self) -> Result<Val, ToValError> {
        Ok(ValE::Array(SlimBoxedSlice::from_vec(self.vec)).into())
    }
}

impl serde::ser::SerializeTuple for SerializeVec {
    type Ok = Val;
    type Error = ToValError;

    fn serialize_element<T>(&mut self, value: &T) -> Result<(), ToValError>
    where
        T: ?Sized + Serialize,
    {
        serde::ser::SerializeSeq::serialize_element(self, value)
    }

    fn end(self) -> Result<Val, ToValError> {
        serde::ser::SerializeSeq::end(self)
    }
}

impl serde::ser::SerializeTupleStruct for SerializeVec {
    type Ok = Val;
    type Error = ToValError;

    fn serialize_field<T>(&mut self, value: &T) -> Result<(), ToValError>
    where
        T: ?Sized + Serialize,
    {
        serde::ser::SerializeSeq::serialize_element(self, value)
    }

    fn end(self) -> Result<Val, ToValError> {
        serde::ser::SerializeSeq::end(self)
    }
}

impl serde::ser::SerializeTupleVariant for SerializeTupleVariant {
    type Ok = Val;
    type Error = ToValError;

    fn serialize_field<T>(&mut self, value: &T) -> Result<(), ToValError>
    where
        T: ?Sized + Serialize,
    {
        self.vec.push(to_val(value)?);
        Ok(())
    }

    fn end(self) -> Result<Val, ToValError> {
        Ok(ValE::Object(Map::new(vec![(
            KeyE::string(self.name).into(),
            ValE::Array(SlimBoxedSlice::from_vec(self.vec)).into(),
        )]))
        .into())
    }
}

impl serde::ser::SerializeMap for SerializeMap {
    type Ok = Val;
    type Error = ToValError;

    fn serialize_key<T>(&mut self, key: &T) -> Result<(), ToValError>
    where
        T: ?Sized + Serialize,
    {
        self.next_key = Some(key.serialize(map_key_serializer::MapKeySerializer)?);
        Ok(())
    }

    fn serialize_value<T>(&mut self, value: &T) -> Result<(), ToValError>
    where
        T: ?Sized + Serialize,
    {
        let key = self.next_key.take();
        // Panic because this indicates a bug in the program rather than an
        // expected failure.
        let key = key.expect("serialize_value called before serialize_key");
        self.map.push((KeyE::string(key).into(), to_val(value)?));
        Ok(())
    }

    fn end(self) -> Result<Val, ToValError> {
        Ok(ValE::Object(Map::new(self.map)).into())
    }
}

mod map_key_serializer;

impl serde::ser::SerializeStruct for SerializeMap {
    type Ok = Val;
    type Error = ToValError;

    fn serialize_field<T>(&mut self, key: &'static str, value: &T) -> Result<(), ToValError>
    where
        T: ?Sized + Serialize,
    {
        serde::ser::SerializeMap::serialize_entry(self, key, value)
    }

    fn end(self) -> Result<Val, ToValError> {
        serde::ser::SerializeMap::end(self)
    }
}

impl serde::ser::SerializeStructVariant for SerializeStructVariant {
    type Ok = Val;
    type Error = ToValError;

    fn serialize_field<T>(&mut self, key: &'static str, value: &T) -> Result<(), ToValError>
    where
        T: ?Sized + Serialize,
    {
        self.map.push((KeyE::str(key).into(), to_val(value)?));
        Ok(())
    }

    fn end(self) -> Result<Val, ToValError> {
        Ok(ValE::Object(Map::new(vec![(
            KeyE::string(self.name).into(),
            ValE::Object(Map::new(self.map)).into(),
        )]))
        .into())
    }
}

impl<'de> Deserialize<'de> for Val {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        deserializer.deserialize_any(ValVisitor)
    }
}

struct ValVisitor;

impl<'de> Visitor<'de> for ValVisitor {
    type Value = Val;

    fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(formatter, "Any valid JSON")
    }

    #[inline]
    fn visit_bool<E>(self, v: bool) -> Result<Self::Value, E>
    where
        E: serde::de::Error,
    {
        Ok(ValE::Bool(v).into())
    }

    #[inline]
    fn visit_i64<E>(self, v: i64) -> Result<Self::Value, E>
    where
        E: serde::de::Error,
    {
        Ok(ValE::Number(NotNan::new(v as f64).unwrap()).into())
    }

    #[inline]
    fn visit_u64<E>(self, v: u64) -> Result<Self::Value, E>
    where
        E: serde::de::Error,
    {
        Ok(ValE::Number(NotNan::new(v as f64).unwrap()).into())
    }

    #[inline]
    fn visit_f64<E>(self, v: f64) -> Result<Self::Value, E>
    where
        E: serde::de::Error,
    {
        Ok(ValE::Number(NotNan::new(v).unwrap()).into())
    }

    #[inline]
    fn visit_str<E>(self, v: &str) -> Result<Self::Value, E>
    where
        E: serde::de::Error,
    {
        Ok(ValE::str(v).into())
    }

    #[inline]
    fn visit_string<E>(self, v: String) -> Result<Self::Value, E>
    where
        E: serde::de::Error,
    {
        Ok(ValE::string(v).into())
    }

    #[inline]
    fn visit_bytes<E>(self, v: &[u8]) -> Result<Self::Value, E>
    where
        E: serde::de::Error,
    {
        let vec = v
            .iter()
            .map(|&b| ValE::Number(NotNan::new(b as f64).unwrap()).into())
            .collect();
        Ok(ValE::Array(SlimBoxedSlice::from_vec(vec)).into())
    }

    #[inline]
    fn visit_none<E>(self) -> Result<Self::Value, E>
    where
        E: serde::de::Error,
    {
        Ok(ValE::Null.into())
    }

    #[inline]
    fn visit_some<D>(self, deserializer: D) -> Result<Self::Value, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        Deserialize::deserialize(deserializer)
    }

    #[inline]
    fn visit_unit<E>(self) -> Result<Self::Value, E>
    where
        E: serde::de::Error,
    {
        Ok(ValE::Null.into())
    }

    #[inline]
    fn visit_newtype_struct<D>(self, deserializer: D) -> Result<Self::Value, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        Deserialize::deserialize(deserializer)
    }

    #[inline]
    fn visit_seq<A>(self, mut seq: A) -> Result<Self::Value, A::Error>
    where
        A: serde::de::SeqAccess<'de>,
    {
        // TODO: proper capacity for SlimBoxedSlice
        let mut vec = Vec::new();

        while let Some(value) = seq.next_element()? {
            vec.push(value);
        }

        Ok(ValE::Array(SlimBoxedSlice::from_vec(vec)).into())
    }

    #[inline]
    fn visit_map<A>(self, mut map: A) -> Result<Self::Value, A::Error>
    where
        A: serde::de::MapAccess<'de>,
    {
        let mut vec = Vec::new();

        while let Some((key, value)) = map.next_entry()? {
            vec.push((key, value));
        }

        Ok(ValE::Object(Map::new(vec)).into())
    }
}

impl<'de> Deserialize<'de> for Key {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        let val = deserializer.deserialize_any(ValVisitor)?;
        Key::try_from(val).map_err(|e| serde::de::Error::custom(e.to_string()))
    }
}

impl From<serde_json::Value> for ValE {
    #[inline]
    fn from(val: serde_json::Value) -> Self {
        match val {
            serde_json::Value::Null => ValE::Null,
            serde_json::Value::Bool(b) => ValE::Bool(b),
            serde_json::Value::Number(n) => ValE::Number(NotNan::new(n.as_f64().unwrap()).unwrap()),
            serde_json::Value::String(s) => ValE::string(s),
            serde_json::Value::Array(a) => ValE::Array(SlimBoxedSlice::from_vec(
                a.into_iter().map(Val::from).collect(),
            )),
            serde_json::Value::Object(o) => ValE::Object(Map::new(
                o.into_iter()
                    .map(|(k, v)| (KeyE::string(k).into(), Val::from(v)))
                    .collect(),
            )),
        }
    }
}

impl From<serde_json::Value> for Val {
    #[inline]
    fn from(val: serde_json::Value) -> Self {
        ValE::from(val).into()
    }
}
