use serde::{
    de::{DeserializeSeed, EnumAccess, IntoDeserializer, Unexpected, VariantAccess, Visitor},
    Deserialize,
};

use crate::{Val, ValE};

use super::{visit_array, visit_object, ValDeserializerError};

pub(crate) struct EnumDeserializer {
    pub(crate) variant: String,
    pub(crate) value: Option<Val>,
}

impl<'de> EnumAccess<'de> for EnumDeserializer {
    type Error = ValDeserializerError;
    type Variant = VariantDeserializer;

    fn variant_seed<V>(self, seed: V) -> Result<(V::Value, VariantDeserializer), Self::Error>
    where
        V: DeserializeSeed<'de>,
    {
        let variant = self.variant.into_deserializer();
        let visitor = VariantDeserializer { value: self.value };
        seed.deserialize(variant).map(|v| (v, visitor))
    }
}

pub(crate) struct VariantDeserializer {
    pub(crate) value: Option<Val>,
}

impl<'de> VariantAccess<'de> for VariantDeserializer {
    type Error = ValDeserializerError;

    fn unit_variant(self) -> Result<(), Self::Error> {
        match self.value {
            Some(value) => Deserialize::deserialize(value),
            None => Ok(()),
        }
    }

    fn newtype_variant_seed<T>(self, seed: T) -> Result<T::Value, Self::Error>
    where
        T: DeserializeSeed<'de>,
    {
        match self.value {
            Some(value) => seed.deserialize(value),
            None => Err(serde::de::Error::invalid_type(
                Unexpected::UnitVariant,
                &"newtype variant",
            )),
        }
    }

    fn tuple_variant<V>(self, _len: usize, visitor: V) -> Result<V::Value, Self::Error>
    where
        V: Visitor<'de>,
    {
        match self.value.map(Into::into) {
            Some(ValE::Array(v)) => {
                if v.as_ref().is_empty() {
                    visitor.visit_unit()
                } else {
                    visit_array(v.into_vec(), visitor)
                }
            }
            Some(other) => Err(serde::de::Error::invalid_type(
                Unexpected::Other(&format!("{:?}", other)),
                &"tuple variant",
            )),
            None => Err(serde::de::Error::invalid_type(
                Unexpected::UnitVariant,
                &"tuple variant",
            )),
        }
    }

    fn struct_variant<V>(
        self,
        _fields: &'static [&'static str],
        visitor: V,
    ) -> Result<V::Value, Self::Error>
    where
        V: Visitor<'de>,
    {
        match self.value.map(Into::into) {
            Some(ValE::Object(v)) => visit_object(v, visitor),
            Some(other) => Err(serde::de::Error::invalid_type(
                Unexpected::Other(&format!("{:?}", other)),
                &"struct variant",
            )),
            None => Err(serde::de::Error::invalid_type(
                Unexpected::UnitVariant,
                &"struct variant",
            )),
        }
    }
}
