use serde::{
    de::{
        value::{MapDeserializer, SeqDeserializer},
        Error, IntoDeserializer, Unexpected, Visitor,
    },
    forward_to_deserialize_any, Deserializer,
};

use super::ValDeserializerError;
use crate::ValRef;

impl<'de> Deserializer<'de> for ValRef<'de> {
    type Error = ValDeserializerError;

    #[inline]
    fn deserialize_any<V>(self, visitor: V) -> Result<V::Value, Self::Error>
    where
        V: serde::de::Visitor<'de>,
    {
        match self {
            ValRef::Null => visitor.visit_none(),
            ValRef::Bool(b) => visitor.visit_bool(b),
            ValRef::Number(n) => {
                // weird hack for deserializing floats when integers are expected
                // if floats are expected, this should still work!
                let float = n.into_inner();
                if float.fract() == 0.0 {
                    visitor.visit_i64(float as i64)
                } else {
                    visitor.visit_f64(float)
                }
            }
            ValRef::String(s) => visitor.visit_str(s.as_ref()),
            ValRef::Array(a) => {
                visitor.visit_seq(SeqDeserializer::new(a.iter().map(|v| v.direct_ref())))
            }
            ValRef::Object(o) => visitor.visit_map(MapDeserializer::new(
                o.0.iter().map(|(k, v)| (k.as_str(), v.direct_ref())),
            )),
        }
    }

    forward_to_deserialize_any! {
        bool i8 i16 i32 i64 i128 u8 u16 u32 u64 u128 f32 f64 char str string
        bytes byte_buf unit unit_struct newtype_struct seq tuple
        tuple_struct map struct identifier ignored_any
    }

    #[inline]
    fn deserialize_option<V>(self, visitor: V) -> Result<V::Value, Self::Error>
    where
        V: Visitor<'de>,
    {
        match self {
            ValRef::Null => visitor.visit_none(),
            _ => visitor.visit_some(self),
        }
    }

    #[inline]
    fn deserialize_enum<V>(
        self,
        _name: &str,
        _variants: &'static [&'static str],
        visitor: V,
    ) -> Result<V::Value, Self::Error>
    where
        V: Visitor<'de>,
    {
        deserialize_val_ref_enum(self, visitor)
    }
}

#[inline]
pub fn deserialize_val_ref_enum<'de, V: Visitor<'de>>(
    val: ValRef<'de>,
    visitor: V,
) -> Result<V::Value, ValDeserializerError> {
    let (variant, value) = match val {
        ValRef::Object(value) => {
            let mut iter = value.0.iter();
            let (variant, value) = match iter.next() {
                Some((k, v)) => (k.as_str().to_string(), v),
                None => {
                    return Err(ValDeserializerError::invalid_value(
                        Unexpected::Map,
                        &"map with a single key",
                    ));
                }
            };
            // enums are encoded in json as maps with a single key:value pair
            if iter.next().is_some() {
                return Err(ValDeserializerError::invalid_value(
                    Unexpected::Map,
                    &"map with a single key",
                ));
            }
            (variant, Some(value))
        }
        ValRef::String(variant) => (variant.to_string(), None),
        other => {
            return Err(ValDeserializerError::invalid_type(
                Unexpected::Other(&format!("{:?}", other)),
                &"string or map",
            ));
        }
    };

    // TODO: save a (deep) clone in the future by implementing borrowing EnumDeserializer
    visitor.visit_enum(super::enum_de::EnumDeserializer {
        variant,
        value: value.cloned(),
    })
}

impl<'de> IntoDeserializer<'de> for ValRef<'de> {
    type Deserializer = Self;

    fn into_deserializer(self) -> Self::Deserializer {
        self
    }
}
